/* $Id$ */
/*-
 * Copyright (c) 2003, 2004 Benedikt Meurer <benny@xfce.org>
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif
#include <stdio.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <libxfce4util/xdg.h>

#include <xprefs/xprefs-error.h>
#include <xprefs/xprefs-private.h>
#include <xprefs-daemon/xprefs-database.h>
#include <xprefs-daemon/xprefs-defaults.h>
#include <xprefs-daemon/xprefs-dir.h>
#include <xprefs-daemon/xprefs-tree.h>


typedef struct
{
  gchar *object;
  gchar *service;
} Listener;

typedef struct
{
  gchar *section;
  GList *listeners;
} SectionWatch;


struct _XPrefsDatabase
{
  DBusConnection *connection;
  XPrefsDefaults *defaults;
  XPrefsTree     *tree;

  guint           scheduled_sync_id;

  GHashTable     *watchers;
};


/* timeout for scheduled syncs (in ms) */
#define SCHEDULED_SYNC_TIMEOUT  (5 * 1000)


static gboolean
xprefs_dbus_message_get_args (DBusConnection *connection,
                              DBusMessage *message,
                              int first_arg_type,
                              ...)
{
  DBusMessage *reply;
  gboolean result;
  va_list valist;

  va_start (valist, first_arg_type);
  result = dbus_message_get_args_valist (message, NULL, first_arg_type, valist);
  va_end (valist);

  if (!result)
    {
      reply = dbus_message_new_error (message, XPREFS_DBUS_ERROR_FAILED,
                                      _("Received a malformed message"));
      dbus_connection_send (connection, reply, NULL);
      dbus_message_unref (reply);
    }

  return result;
}

static gboolean
xprefs_dbus_set_gerror (DBusConnection *connection,
                        DBusMessage *message,
                        GError **error)
{
  const gchar *ename = NULL;
  DBusMessage *reply;
  XPrefsError ecode;

  if (error == NULL || *error == NULL)
    return FALSE;

  ecode = (*error)->code;

  /* success is not supposed to get set (muha, I like this comment :-) */
  g_return_val_if_fail (ecode != XPREFS_ERROR_SUCCESS, FALSE);

  switch (ecode)
    {
    case XPREFS_ERROR_FAILED:
      ename = XPREFS_DBUS_ERROR_FAILED;
      break;

    case XPREFS_ERROR_INVALID_VALUE:
      ename = XPREFS_DBUS_ERROR_INVALID_VALUE;
      break;

    case XPREFS_ERROR_TYPE_MISMATCH:
      ename = XPREFS_DBUS_ERROR_TYPE_MISMATCH;
      break;

    case XPREFS_ERROR_BAD_KEY:
      ename = XPREFS_DBUS_ERROR_BAD_KEY;
      break;

    case XPREFS_ERROR_IS_DIR:
      ename = XPREFS_DBUS_ERROR_IS_DIR;
      break;

    case XPREFS_ERROR_IS_KEY:
      ename = XPREFS_DBUS_ERROR_IS_KEY;
      break;

    default:
      g_assert_not_reached ();
      break;
    }

  reply = dbus_message_new_error (message, ename, (*error)->message);
  dbus_connection_send (connection, reply, NULL);
  dbus_message_unref (reply);

  g_error_free (*error);
  *error = NULL;

  return TRUE;
}


static gchar*
xprefs_database_path (const gchar *filename)
{
  gchar *directory;
  gchar *path;

  directory = xfce_xdg_save_config_path ("xprefs");
  path = g_build_filename (directory, filename, NULL);
  g_free (directory);

  return path;
}


static void
xprefs_database_handle_get (XPrefsDatabase *database,
                            DBusMessage *message)
{
  const XPrefsValue *value;
  dbus_uint32_t flags;
  DBusMessage *reply;
  GError *gerror;
  gchar *key;

  gerror = NULL;

  if (!xprefs_dbus_message_get_args (database->connection, message, 
                                     DBUS_TYPE_STRING, &key,
                                     DBUS_TYPE_UINT32, &flags,
                                     DBUS_TYPE_INVALID))
    return;

  value = xprefs_tree_get (database->tree, key, &gerror);

  /* check for a default setting */
  if (value == NULL)
    {
      value = xprefs_defaults_lookup (database->defaults, key);
      
      if (value != NULL)
        g_clear_error (&gerror);
    }

  dbus_free (key);

  if (xprefs_dbus_set_gerror (database->connection, message, &gerror))
    return;

  reply = dbus_message_new_method_return (message);
  dbus_message_append_xprefs_value (reply, value);
  dbus_connection_send (database->connection, reply, NULL);
  dbus_message_unref (reply);
}

static void
xprefs_database_handle_set (XPrefsDatabase *database,
                            DBusMessage *message)
{
  DBusMessageIter iter;
  dbus_uint32_t flags;
  DBusMessage *reply;
  XPrefsValue *value;
  GError *gerror;
  gchar *key;

  gerror = NULL;

  dbus_message_iter_init (message, &iter);
  key = dbus_message_iter_get_string (&iter);
  dbus_message_iter_next (&iter);
  flags = dbus_message_iter_get_uint32 (&iter);
  dbus_message_iter_next (&iter);
  value = dbus_message_iter_get_xprefs_value (&iter);

  xprefs_tree_set (database->tree, key, value, &gerror);

  if (xprefs_dbus_set_gerror (database->connection, message, &gerror))
    {
      dbus_free (key);
      xprefs_value_free (value);
      return;
    }

  /* schedule a database sync */
  xprefs_database_schedule_sync (database);

  /* notify listeners */
  xprefs_database_notify (database, key, value);
  dbus_free (key);

  reply = dbus_message_new_method_return (message);
  dbus_connection_send (database->connection, reply, NULL);
  dbus_message_unref (reply);
}

static void
xprefs_database_handle_unset (XPrefsDatabase *database,
                              DBusMessage *message)
{
  const XPrefsValue *value;
  dbus_uint32_t flags;
  DBusMessage *reply;
  GError *gerror;
  gchar *key;

  gerror = NULL;

  if (!xprefs_dbus_message_get_args (database->connection, message, 
                                     DBUS_TYPE_STRING, &key,
                                     DBUS_TYPE_UINT32, &flags,
                                     DBUS_TYPE_INVALID))
    return;

  xprefs_tree_unset (database->tree, key, &gerror);

  if (xprefs_dbus_set_gerror (database->connection, message, &gerror))
    {
      dbus_free (key);
      return;
    }

  /* schedule a database sync */
  xprefs_database_schedule_sync (database);

  /* notify listeners (using default value if any) */
  value = xprefs_defaults_lookup (database->defaults, key);
  xprefs_database_notify (database, key, value);
  dbus_free (key);

  reply = dbus_message_new_method_return (message);
  dbus_connection_send (database->connection, reply, NULL);
  dbus_message_unref (reply);
}

static void
xprefs_database_handle_dir_exists (XPrefsDatabase *database,
                                   DBusMessage *message)
{
  dbus_uint32_t flags;
  dbus_bool_t result;
  DBusMessage *reply;
  GError *gerror;
  gchar *key;

  gerror = NULL;

  if (!xprefs_dbus_message_get_args (database->connection, message,
                                     DBUS_TYPE_STRING, &key,
                                     DBUS_TYPE_UINT32, &flags,
                                     DBUS_TYPE_INVALID))
    return;

  result = xprefs_tree_dir_exists (database->tree, key, &gerror);
  dbus_free (key);
  
  if (xprefs_dbus_set_gerror (database->connection, message, &gerror))
    return;

  reply = dbus_message_new_method_return (message);
  dbus_message_append_args (message,
                            DBUS_TYPE_BOOLEAN, result,
                            DBUS_TYPE_INVALID);
  dbus_connection_send (database->connection, reply, NULL);
  dbus_message_unref (reply);
}

static void
xprefs_database_handle_all_entries (XPrefsDatabase *database,
                                    DBusMessage *message)
{
  DBusMessageIter iter;
  dbus_uint32_t flags;
  DBusMessage *reply;
  GError *gerror;
  GList *entries;
  GList *entry;
  gchar *key;

  if (!xprefs_dbus_message_get_args (database->connection, message,
                                     DBUS_TYPE_STRING, &key,
                                     DBUS_TYPE_UINT32, &flags,
                                     DBUS_TYPE_INVALID))
    return;

  gerror = NULL;
  entries = xprefs_tree_all_entries (database->tree, key, &gerror);
  dbus_free (key);

  if (xprefs_dbus_set_gerror (database->connection, message, &gerror))
    return;

  reply = dbus_message_new_method_return (message);
  dbus_message_append_iter_init (reply, &iter);
  for (entry = entries; entry != NULL; entry = entry->next)
    {
      dbus_message_iter_append_xprefs_entry (&iter, XPREFS_ENTRY (entry->data));
    }
  dbus_connection_send (database->connection, reply, NULL);
  dbus_message_unref (reply);
}

static void
xprefs_database_handle_all_subdirs (XPrefsDatabase *database,
                                    DBusMessage *message)
{
  DBusMessageIter iter;
  dbus_uint32_t flags;
  DBusMessage *reply;
  const gchar *name;
  GError *gerror;
  GList *subdirs;
  GList *subdir;
  gchar *key;

  if (!xprefs_dbus_message_get_args (database->connection, message,
                                     DBUS_TYPE_STRING, &key,
                                     DBUS_TYPE_UINT32, &flags,
                                     DBUS_TYPE_INVALID))
    return;

  gerror = NULL;
  subdirs = xprefs_tree_all_subdirs (database->tree, key, &gerror);
  dbus_free (key);

  if (xprefs_dbus_set_gerror (database->connection, message, &gerror))
    return;

  reply = dbus_message_new_method_return (message);
  dbus_message_append_iter_init (reply, &iter);
  for (subdir = subdirs; subdir != NULL; subdir = subdir->next)
    {
      name = xprefs_dir_get_name (XPREFS_DIR (subdir->data));
      dbus_message_iter_append_string (&iter, name);
    }
  dbus_connection_send (database->connection, reply, NULL);
  dbus_message_unref (reply);
}

static void
xprefs_database_handle_listener_add (XPrefsDatabase *database,
                                     DBusMessage *message)
{
  SectionWatch *watch;
  Listener *listener;
  DBusMessage *reply;
  dbus_uint32_t flags;
  const gchar *sender;
  gchar *namespace_section;
  gchar *object_path;

  if (!xprefs_dbus_message_get_args (database->connection, message,
                                     DBUS_TYPE_STRING, &namespace_section,
                                     DBUS_TYPE_UINT32, &flags,
                                     DBUS_TYPE_STRING, &object_path,
                                     DBUS_TYPE_INVALID))
    return;

  sender = dbus_message_get_sender (message);

  watch = g_hash_table_lookup (database->watchers, namespace_section);
  if (watch != NULL)
    {
      watch = g_new (SectionWatch, 1);
      watch->section = g_strdup (namespace_section);
      watch->listeners = NULL;

      g_hash_table_insert (database->watchers, watch->section, watch);
    }

  listener = g_new (Listener, 1);
  listener->service = g_strdup (sender);
  listener->object = g_strdup (object_path);

  watch->listeners = g_list_prepend (watch->listeners, listener);

  dbus_free (namespace_section);
  dbus_free (object_path);
  
  reply = dbus_message_new_method_return (message);
  dbus_connection_send (database->connection, reply, NULL);
  dbus_message_unref (reply);
}

static void
xprefs_database_handle_listener_remove (XPrefsDatabase *database,
                                        DBusMessage *message)
{
  Listener *listener;
  SectionWatch *watch;
  const gchar *sender;
  DBusMessage *reply;
  dbus_uint32_t flags;
  gchar *namespace_section;
  gchar *object_path;
  GList *l;

  if (!xprefs_dbus_message_get_args (database->connection, message,
                                     DBUS_TYPE_STRING, &namespace_section,
                                     DBUS_TYPE_UINT32, &flags,
                                     DBUS_TYPE_STRING, &object_path,
                                     DBUS_TYPE_INVALID))
    return;

  sender = dbus_message_get_sender (message);

  watch = g_hash_table_lookup (database->watchers, namespace_section);
  if (watch != NULL)
    {
      for (l = watch->listeners; l != NULL; l = l->next)
        {
          listener = (Listener *) l->data;

          if (strcmp (listener->service, sender) == 0
              && strcmp (listener->object, object_path) == 0)
            break;
        }

      if (l != NULL)
        {
          watch->listeners = g_list_remove_link (watch->listeners, l);

          /* if the watch contains no more listeners, free it */
          if (g_list_length (watch->listeners) == 0)
            {
              g_hash_table_remove (database->watchers, watch->section);
              g_free (watch->section);
              g_free (watch);
            }

          /* free the listener */
          g_free (listener->object);
          g_free (listener->service);
          g_list_free_1 (l);
        }
    }

  dbus_free (namespace_section);
  dbus_free (object_path);

  reply = dbus_message_new_method_return (message);
  dbus_connection_send (database->connection, reply, NULL);
  dbus_message_unref (reply);
}


static void
handle_unregister (DBusConnection *dbus_connection, void *user_data)
{
}

static DBusHandlerResult
handle_message (DBusConnection *connection,
                DBusMessage *message,
                void *user_data)
{
  XPrefsDatabase *database = (XPrefsDatabase *) user_data;

  if (dbus_message_is_signal (message,
                              DBUS_INTERFACE_ORG_FREEDESKTOP_LOCAL,
                              "Disconnected"))
    {
      dbus_connection_unref (database->connection);
      database->connection = NULL;
      /* XXX - log message and exit main loop, will sync before exit */
    }
  else if (dbus_message_is_method_call (message,
                                        XPREFS_DBUS_DATABASE_INTERFACE,
                                        XPREFS_DBUS_DATABASE_GET))
    {
      xprefs_database_handle_get (database, message);
    }
  else if (dbus_message_is_method_call (message,
                                        XPREFS_DBUS_DATABASE_INTERFACE,
                                        XPREFS_DBUS_DATABASE_SET))
    {
      xprefs_database_handle_set (database, message);
    }
  else if (dbus_message_is_method_call (message,
                                        XPREFS_DBUS_DATABASE_INTERFACE,
                                        XPREFS_DBUS_DATABASE_UNSET))
    {
      xprefs_database_handle_unset (database, message);
    }
  else if (dbus_message_is_method_call (message,
                                        XPREFS_DBUS_DATABASE_INTERFACE,
                                        XPREFS_DBUS_DATABASE_DIR_EXISTS))
    {
      xprefs_database_handle_dir_exists (database, message);
    }
  else if (dbus_message_is_method_call (message,
                                        XPREFS_DBUS_DATABASE_INTERFACE,
                                        XPREFS_DBUS_DATABASE_ALL_ENTRIES))
    {
      xprefs_database_handle_all_entries (database, message);
    }
  else if (dbus_message_is_method_call (message,
                                        XPREFS_DBUS_DATABASE_INTERFACE,
                                        XPREFS_DBUS_DATABASE_ALL_SUBDIRS))
    {
      xprefs_database_handle_all_subdirs (database, message);
    }
  else if (dbus_message_is_method_call (message,
                                        XPREFS_DBUS_DATABASE_INTERFACE,
                                        XPREFS_DBUS_DATABASE_LISTENER_ADD))
    {
      xprefs_database_handle_listener_add (database, message);
    }
  else if (dbus_message_is_method_call (message,
                                        XPREFS_DBUS_DATABASE_INTERFACE,
                                        XPREFS_DBUS_DATABASE_LISTENER_REMOVE))
    {
      xprefs_database_handle_listener_remove (database, message);
    }
  else
    {
      return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
    }

  return DBUS_HANDLER_RESULT_HANDLED;
}

static const struct DBusObjectPathVTable handler_vtable = 
{
  handle_unregister,
  handle_message,
  NULL,
};


XPrefsDatabase*
xprefs_database_new (DBusConnection *connection)
{
  XPrefsDatabase *database;
  gchar *filename;
  gchar **path;

  database = g_new0 (XPrefsDatabase, 1);
  database->connection = connection;
  database->tree = xprefs_tree_new ();
  database->defaults = xprefs_defaults_new ();
  database->watchers = g_hash_table_new (g_str_hash, g_str_equal);

  dbus_connection_ref (connection);

  /* load saved state */
  filename = xprefs_database_path ("database.xml");
  xprefs_database_parse_file (database, filename);
  g_free (filename);

  /* watch out for ServiceDeleted signals from the bus daemon */
  dbus_bus_add_match (connection,
                      "sender=org.freedesktop.DBus,member=ServiceDeleted",
                      NULL);

  /* register object path */
  path = g_strsplit (XPREFS_DBUS_DATABASE_OBJECT + 1, "/", -1);
  dbus_connection_register_object_path (connection,
                                        (const gchar **) path,
                                        &handler_vtable,
                                        database);
  g_strfreev (path);

  dbus_bus_acquire_service (connection,
                            XPREFS_DBUS_SERVICE,
                            0,
                            NULL);

  return database;
}

void
xprefs_database_free (XPrefsDatabase *database)
{
}

static const gchar*
xprefs_value_to_xml (const XPrefsValue *value)
{
  static gchar buffer[2048];
  gchar *string_escaped;
  gchar *string_raw;

  string_raw = xprefs_value_to_string (value);
  string_escaped = g_markup_escape_text (string_raw, strlen (string_raw));

  g_snprintf (buffer, sizeof (buffer), "<value type=\"%s\">%s</value>",
              xprefs_value_type_to_string (value->type), string_escaped);

  g_free (string_escaped);
  g_free (string_raw);

  return buffer;
}

static void
print_entry (XPrefsEntry *entry, gpointer user_data)
{
  const XPrefsValue *value;
  const gchar *name;
  XPrefsDir *dir;
  gchar *path;
  FILE *fp;

  fp = (FILE *) user_data;
  dir = XPREFS_DIR (xprefs_entry_get_user_data (entry));
  path = xprefs_dir_absolute_path (dir);
  value = xprefs_entry_get_value (entry);
  name = xprefs_entry_get_name (entry);

  fprintf (fp,
           "  <xprefs:entry>\n"
           "    <key>%s%s</key>\n"
           "    %s\n"
           "  </xprefs:entry>\n"
           "\n",
           path, name,
           xprefs_value_to_xml (value));

  g_free (path);
}

void
xprefs_database_sync (XPrefsDatabase *database)
{
  gchar *filename;
  gchar *realname;
  FILE *fp;

  /* build up database filename */
  filename = xprefs_database_path (".database.xml");
  realname = xprefs_database_path ("database.xml");

  fp = fopen (filename, "wb");
  if (fp == NULL)
    {
      g_printerr ("Unable to open database file %s.\n", filename);
      g_free (filename);
      g_free (realname);
      return;
    }

  /* print header */
  fprintf (fp,
           "<?xml encoding=\"UTF-8\" version=\"1.0\"?>\n"
           "<!DOCTYPE xprefs:database>\n"
           "<xprefs:database>\n"
           "\n");

  /* print entries */
  xprefs_tree_foreach (database->tree,
                       print_entry,
                       (gpointer) fp);

  /* print footer */
  fprintf (fp,
           "</xprefs:database>\n");

  fclose (fp);

  /* now the fun with linking :) */
  unlink (realname);
  link (filename, realname);
  unlink (filename);

  g_free (filename);
  g_free (realname);
}

static gboolean
scheduled_sync_callback (gpointer user_data)
{
  XPrefsDatabase *database = (XPrefsDatabase *) user_data;

  xprefs_database_sync (database);
  database->scheduled_sync_id = 0;
  
  /* remove the timer source */
  return FALSE;
}

/**
 * xprefs_database_schedule_sync:
 * @database:
 */
void
xprefs_database_schedule_sync (XPrefsDatabase *database)
{
  /* remove prior established timer */
  if (database->scheduled_sync_id != 0)
    g_source_remove (database->scheduled_sync_id);

  /* re-add timer */
  database->scheduled_sync_id = g_timeout_add (SCHEDULED_SYNC_TIMEOUT,
                                               scheduled_sync_callback,
                                               database);
}


/**
 * xprefs_database_notify:
 * @database:
 * @key:
 * @value:
 */
void
xprefs_database_notify (XPrefsDatabase    *database,
                        const gchar       *key,
                        const XPrefsValue *value)
{
  SectionWatch *watch;
  DBusMessageIter iter;
  DBusMessage *message;
  XPrefsEntry *entry;
  gchar *section;
  gchar *sep;
  GList *l;

  section = g_strdup (key);
  entry = xprefs_entry_new (key, value);

  message = dbus_message_new (DBUS_MESSAGE_TYPE_METHOD_CALL);
  dbus_message_set_interface (message, XPREFS_DBUS_LISTENER_INTERFACE);
  dbus_message_set_member (message, XPREFS_DBUS_LISTENER_NOTIFY);
  dbus_message_set_no_reply (message, TRUE);
  dbus_message_append_iter_init (message, &iter);
  dbus_message_iter_append_xprefs_entry (&iter, entry);

  for (;;)
    {
      watch = g_hash_table_lookup (database->watchers, section);
      if (watch != NULL)
        {
          for (l = watch->listeners; l != NULL; l = l->next)
            {
              Listener *listener = (Listener *) l->data;

              dbus_message_set_destination (message, listener->service);
              dbus_message_set_path (message, listener->object);

              dbus_connection_send (database->connection, message, NULL);
            }
        }

      /* remove the trailing leaf */
      sep = strrchr (section, '/');
      if (sep == section)
        break;
      *sep = '\0';
    }
  
  dbus_message_unref (message);

  xprefs_entry_free (entry);
  g_free (section);
}


static void   start_element_handler     (GMarkupParseContext *context,
                                         const gchar         *element_name,
                                         const gchar        **attribute_names,
                                         const gchar        **attribute_values,
                                         gpointer             user_data,
                                         GError             **error);
static void   end_element_handler       (GMarkupParseContext *context,
                                         const gchar         *element_name,
                                         gpointer             user_data,
                                         GError             **error);
static void   text_handler              (GMarkupParseContext *context,
                                         const gchar         *text,
                                         gsize                text_len,  
                                         gpointer             user_data,
                                         GError             **error);
static void   error_handler             (GMarkupParseContext *context,
                                         GError              *error,
                                         gpointer             user_data);


typedef enum _ParserState ParserState;
enum _ParserState
{
  PARSER_START,
  PARSER_DATABASE,
  PARSER_ENTRY,
  PARSER_KEY,
  PARSER_VALUE,
};

typedef struct _Parser Parser;
struct _Parser
{
  ParserState       state;
  XPrefsDatabase   *database;
  const gchar      *filename;

  gchar            *key;
  XPrefsValue      *value;
  XPrefsValueType   value_type;
};


static void
start_element_handler (GMarkupParseContext *context,
                       const gchar         *element_name,
                       const gchar        **attribute_names,
                       const gchar        **attribute_values,
                       gpointer             user_data,
                       GError             **error)
{
  Parser *parser;
  int i;

  parser = (Parser *) user_data;

  switch (parser->state)
    {
    case PARSER_START:
      if (strcmp (element_name, "xprefs:database") == 0)
        parser->state = PARSER_DATABASE;
      break;

    case PARSER_DATABASE:
      if (strcmp (element_name, "xprefs:entry") == 0)
        parser->state = PARSER_ENTRY;
      break;

    case PARSER_ENTRY:
      if (strcmp (element_name, "key") == 0)
        parser->state = PARSER_KEY;
      else if (strcmp (element_name, "value") == 0)
        {
          /* look for the type */
          parser->value_type = XPREFS_VALUE_INVALID;
          for (i = 0; attribute_names[i] != NULL; ++i)
            if (strcmp (attribute_names[i], "type") == 0)
              {
                parser->value_type =
                  xprefs_value_type_from_string (attribute_values[i]);
              }

          if (parser->value_type != XPREFS_VALUE_INVALID)
            parser->state = PARSER_VALUE;
        }
      break;

    default:
      /* XXX */
      break;
    }
}

static void
end_element_handler (GMarkupParseContext *context,
                     const gchar         *element_name,
                     gpointer             user_data,
                     GError             **error)
{
  Parser *parser;
  
  parser = (Parser *) user_data;

  switch (parser->state)
    {
    case PARSER_START:
      g_warning ("parser: This shouldn't happen.");
      break;

    case PARSER_DATABASE:
      if (strcmp (element_name, "xprefs:database") == 0)
        parser->state = PARSER_START;
      break;

    case PARSER_ENTRY:
      if (strcmp (element_name, "xprefs:entry") == 0)
        {
          if (parser->key != NULL)
            {
              xprefs_tree_set (parser->database->tree,
                               parser->key,
                               parser->value,
                               NULL);

              g_free (parser->key);
            }
          else if (parser->value != NULL)
            {
              xprefs_value_free (parser->value);
            }

          parser->key = NULL;
          parser->value = NULL;
          parser->state = PARSER_DATABASE;
        }
      break;

    case PARSER_KEY:
      if (strcmp (element_name, "key") == 0)
        parser->state = PARSER_ENTRY;
      break;

    case PARSER_VALUE:
      if (strcmp (element_name, "value") == 0)
        parser->state = PARSER_ENTRY;
      break;
    }
}

static void
text_handler (GMarkupParseContext *context,
              const gchar         *text,
              gsize                text_len,  
              gpointer             user_data,
              GError             **error)
{
  Parser *parser;
  
  parser = (Parser *) user_data;

  switch (parser->state)
    {
    case PARSER_KEY:
      if (parser->key != NULL)
        g_free (parser->key);
      parser->key = g_strdup (text);
      break;

    case PARSER_VALUE:
      if (parser->value != NULL)
        xprefs_value_free (parser->value);
      parser->value = xprefs_value_new_from_string (parser->value_type,
                                                    text,
                                                    error);
      break;

    default:
      /* XXX */
      break;
    }
}

static void
error_handler (GMarkupParseContext *context,
               GError              *error,
               gpointer             user_data)
{
  if (error != NULL && error->message != NULL)
    g_warning (" %s", error->message);
}

/**
 * xprefs_database_parse_file:
 * @database:
 * @filename:
 **/
void
xprefs_database_parse_file (XPrefsDatabase *database,
                            const gchar *filename)
{
  static GMarkupParser markup_parser = {
    start_element_handler,
    end_element_handler,
    text_handler,
    NULL,
    error_handler,
  };

  GMarkupParseContext *context;
  gchar *contents = NULL;
  Parser parser;

  if (!g_file_get_contents (filename, &contents, NULL, NULL))
    return;

  parser.state = PARSER_START;
  parser.database = database;
  parser.filename = filename;
  parser.key = NULL;
  parser.value = NULL;
  parser.value_type = XPREFS_VALUE_INVALID;

  context = g_markup_parse_context_new (&markup_parser, 0, &parser, NULL);
  g_markup_parse_context_parse (context, contents, strlen (contents), NULL);
  g_markup_parse_context_end_parse (context, NULL);
  g_markup_parse_context_free (context);
  
  g_free (contents);
}


