/* $Id$ */
/*-
 * Copyright (c) 2003, 2004 Benedikt Meurer <benny@xfce.org>
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <xprefs/xprefs-error.h>
#include <xprefs/xprefs-private.h>
#include <xprefs-daemon/xprefs-dir.h>


struct _XPrefsDir
{
  gchar     *name;
  XPrefsDir *parent;
  GList     *entries;
  GList     *subdirs;
  GList     *listeners;
};


static void
xprefs_set_error_is_dir (GError **error,
                         XPrefsDir *parent,
                         const gchar *name)
{
  gchar *path = xprefs_dir_absolute_path (parent);
  g_set_error (error, XPREFS_ERROR, XPREFS_ERROR_IS_DIR,
               _("'%s%s' is a directory"), path, name);
  g_free (path);
}

static void
xprefs_set_error_is_key (GError **error,
                         XPrefsDir *parent,
                         const gchar *name)
{
  gchar *path = xprefs_dir_absolute_path (parent);
  g_set_error (error, XPREFS_ERROR, XPREFS_ERROR_IS_KEY,
               _("'%s%s' is a key entry"), path, name);
  g_free (path);
}

static XPrefsEntry*
xprefs_dir_add_entry (XPrefsDir *dir,
                      const gchar *name,
                      GError **error)
{
  XPrefsEntry *entry;

  /* verify that there's no entry of that name */
  if (G_UNLIKELY (xprefs_dir_get_subdir (dir, name) != NULL))
    {
      xprefs_set_error_is_dir (error, dir, name);
      return NULL;
    }

  /* create new entry node */
  entry = xprefs_entry_new_nocopy (g_strdup (name), NULL);
  xprefs_entry_set_user_data (entry, dir, NULL);

  /* link with parent node */
  dir->entries = g_list_append (dir->entries, entry);

  return entry;
}


/**
 * xprefs_dir_new:
 * @parent:
 * @name:
 * @error:
 *
 * Return value: the new XPrefsDir object or NULL on failure.
 **/
XPrefsDir*
xprefs_dir_new (XPrefsDir *parent,
                const gchar *name,
                GError **error)
{
  XPrefsDir *dir;

  /* verify that there's no entry of that name */
  if (parent && G_UNLIKELY (xprefs_dir_get_entry (parent, name) != NULL))
    {
      xprefs_set_error_is_key (error, parent, name);
      return NULL;
    }

  /* create new directory node */
  dir = g_new0 (XPrefsDir, 1);
  dir->name = g_strdup (name);
  dir->parent = parent;

  /* link with parent node (if any) */
  if (G_LIKELY (parent != NULL))
    parent->subdirs = g_list_append (parent->subdirs, dir);

  return dir;
}

/**
 * xprefs_dir_free:
 * @dir: XPrefsDir object to free.
 *
 * Unlinks @dir from the parent directory node and frees the resources
 * covered by the directory node.
 *
 * XXX - What about the subdirs and entries lists?
 */
void
xprefs_dir_free (XPrefsDir *dir)
{
  XPrefsDir *parent = dir->parent;

  /* unlink from parent node (if any) */
  if (G_LIKELY (parent != NULL))
    parent->subdirs = g_list_remove (parent->subdirs, dir);

  /* free directory node data */
  g_free (dir->name);
  g_free (dir);
}

/**
 * xprefs_node_absolute_path:
 * @dir:
 *
 * Return value: the absolute path to @dir.
 **/
gchar*
xprefs_dir_absolute_path (const XPrefsDir *dir)
{
  gchar *parent_path;
  gchar *path;

  if (dir->parent == NULL)
    {
      path = g_strdup ("/");
    }
  else
    {
      parent_path = xprefs_dir_absolute_path (dir->parent);
      path = g_strconcat (parent_path, dir->name, "/", NULL);
      g_free (parent_path);
    }

  return path;
}

/**
 * xprefs_dir_get_name:
 * @dir:
 *
 * Return value: the directories name.
 **/
const gchar*
xprefs_dir_get_name (const XPrefsDir *dir)
{
  return dir->name;
}

/**
 * xprefs_dir_get_entries:
 * @dir:
 *
 * Return value: list of all entries in @dir.
 **/
GList*
xprefs_dir_get_entries (const XPrefsDir *dir)
{
  return dir->entries;
}

/**
 * xprefs_dir_get_subdirs:
 * @dir:
 *
 * Return value: list of all subdirs in @dir.
 **/
GList*
xprefs_dir_get_subdirs (const XPrefsDir *dir)
{
  return dir->subdirs;
}

/**
 * xprefs_dir_get_entry:
 * @dir:
 * @name:
 *
 * Return value: the entry named @name or NULL if not found.
 **/
XPrefsEntry *
xprefs_dir_get_entry (const XPrefsDir *dir, const gchar *name)
{
  GList *l;

  for (l = dir->entries; l != NULL; l = l->next)
    if (strcmp (xprefs_entry_get_name (XPREFS_ENTRY (l->data)), name) == 0)
      return XPREFS_ENTRY (l->data);

  return NULL;
}

/**
 * xprefs_dir_get_subdir:
 * @dir:
 * @name: name of the sub directory to lookup.
 *
 * Return value: the sub directory named @name or NULL if not found.
 **/
XPrefsDir *
xprefs_dir_get_subdir (const XPrefsDir *dir, const gchar *name)
{
  GList *l;

  for (l = dir->subdirs; l != NULL; l = l->next)
    if (strcmp (XPREFS_DIR (l->data)->name, name) == 0)
      return XPREFS_DIR (l->data);

  return NULL;
}

/**
 * xprefs_dir_lookup_entry:
 * @dir:
 * @components:
 * @create_if_necessary:
 * @error:
 *
 * Return value: entry or NULL if error or not found.
 **/
XPrefsEntry*
xprefs_dir_lookup_entry (XPrefsDir *dir, 
                         gchar **components,
                         gboolean create_if_necessary,
                         GError **error)
{
  XPrefsEntry *entry;
  XPrefsDir *child;
  guint n;

  for (n = 0; components[n] != NULL; ++n)
    {
      const gchar *component = components[n];

      /* ignore empty path components */
      if (G_UNLIKELY (*component == '\0'))
        continue;

      /* fetch the entry if this is the last path component */
      if (components[n + 1] == NULL)
        {
          entry = xprefs_dir_get_entry (dir, component);

          /* automatically create the missing entry node */
          if (entry == NULL && create_if_necessary)
            {
              entry = xprefs_dir_add_entry (dir, component, error);
            }

          return entry;
        }

      /* change to subdir */
      child = xprefs_dir_get_subdir (dir, component);
      if (child == NULL)
        {
          if (create_if_necessary)
            {
              child = xprefs_dir_new (dir, component, error);
              if (child == NULL)
                break;
            }
          else
            break;
        }
      dir = child;
    }

  return NULL;
}

/**
 * xprefs_dir_lookup_subdir:
 * @dir:
 * @components:
 * @create_if_necessary:
 * @error:
 *
 * Return value: sub directory or NULL if error or not found.
 **/
XPrefsDir*
xprefs_dir_lookup_subdir (XPrefsDir *dir,
                          gchar **components,
                          gboolean create_if_necessary,
                          GError **error)
{
  XPrefsDir *child;
  guint n;

  for (n = 0; components[n] != NULL; ++n)
    {
      const gchar *component = components[n];

      /* ignore empty path components */
      if (G_UNLIKELY (*component == '\0'))
        continue;

      /* change to subdir */
      child = xprefs_dir_get_subdir (dir, component);
      if (child == NULL)
        {
          if (create_if_necessary)
            {
              child = xprefs_dir_new (dir, component, error);
              if (child == NULL)
                break;
            }
          else
            break;
        }
      dir = child;
    }

  return dir;
}

/**
 **/
gboolean
xprefs_dir_remove_entry (XPrefsDir *dir,
                         gchar **components,
                         GError **error)
{
  XPrefsEntry *entry;
  XPrefsDir *parent;

  /* lookup matching entry */
  entry = xprefs_dir_lookup_entry (dir, components, FALSE, error);
  if (entry == NULL)
    return FALSE;

  /* unlink from parent */
  parent = XPREFS_DIR (xprefs_entry_get_user_data (entry));
  parent->entries = g_list_remove (parent->entries, entry);

  /* remove entry data */
  xprefs_entry_free (entry);

  return TRUE;
}


