/* $Id$ */
/*-
 * Copyright (c) 2003, 2004 Benedikt Meurer <benny@xfce.org>
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <xprefs/xprefs-error.h>
#include <xprefs/xprefs-private.h>
#include <xprefs-daemon/xprefs-dir.h>
#include <xprefs-daemon/xprefs-tree.h>

struct _XPrefsTree
{
  XPrefsDir *root;
};


static gchar**
xprefs_split_path (const gchar *path)
{
  while (*path == '/')
    ++path;

  if (*path == '\0')
    return NULL;

  return g_strsplit (path, "/", -1);
}


/**
 * xprefs_tree_new:
 **/
XPrefsTree*
xprefs_tree_new (void)
{
  XPrefsTree *tree;

  tree = g_new0 (XPrefsTree, 1);
  tree->root = xprefs_dir_new (NULL, "/", NULL);

  return tree;
}

/**
 * xprefs_tree_get:
 * @tree:
 * @key:
 * @error:
 *
 * Return value:
 **/
const XPrefsValue*
xprefs_tree_get (XPrefsTree *tree,
                 const gchar *key,
                 GError **error)
{
  XPrefsEntry *entry;
  gchar **components;

  components = xprefs_split_path (key);
  if (G_LIKELY (components != NULL))
    {
      entry = xprefs_dir_lookup_entry (tree->root, components, FALSE, error);
      g_strfreev (components);

      if (G_LIKELY (entry != NULL))
          return xprefs_entry_get_value (entry);
    }

  g_set_error (error, XPREFS_ERROR, XPREFS_ERROR_BAD_KEY,
               _("Path '%s' does not point to a valid entry"),
               key);

  return NULL;
}

gboolean
xprefs_tree_set (XPrefsTree *tree,
                 const gchar *key,
                 XPrefsValue *value,
                 GError **error)
{
  XPrefsEntry *entry;
  gchar **components;

  components = xprefs_split_path (key);
  if (G_LIKELY (components != NULL))
    {
      entry = xprefs_dir_lookup_entry (tree->root, components, TRUE, error);
      g_strfreev (components);

      if (G_LIKELY (entry != NULL))
        {
          xprefs_entry_set_value (entry, value);
          return TRUE;
        }
    }

  /* XXX - I want a better error message here! */
  g_set_error (error, XPREFS_ERROR, XPREFS_ERROR_BAD_KEY,
               _("Unable to set new value for '%s'"), key);

  return FALSE;
}

gboolean
xprefs_tree_unset (XPrefsTree *tree,
                   const gchar *key,
                   GError **error)
{
  gboolean result = FALSE;
  gchar **components;

  components = xprefs_split_path (key);
  if (G_LIKELY (components != NULL))
    {
      result = xprefs_dir_remove_entry (tree->root, components, error);
      g_strfreev (components);

      if (result)
        return TRUE;

      /* FALL-THROUGH */
    }

  g_set_error (error, XPREFS_ERROR, XPREFS_ERROR_BAD_KEY,
               _("Invalid key '%s'"), key);

  return FALSE;
}

gboolean
xprefs_tree_dir_exists (XPrefsTree *tree,
                        const gchar *key,
                        GError **error)
{
  gchar **components;
  XPrefsDir *dir;

  /* treat special case */
  if (G_UNLIKELY (key[0] == '/' && key[1] == '\0'))
    return TRUE;

  components = xprefs_split_path (key);
  if (G_LIKELY (components != NULL))
    {
      dir = xprefs_dir_lookup_subdir (tree->root, components, FALSE, error);
      g_strfreev (components);

      return (dir != NULL);
    }

  g_set_error (error, XPREFS_ERROR, XPREFS_ERROR_BAD_KEY,
               _("Invalid key '%s'"), key);

  return FALSE;
}

/**
 * xprefs_tree_all_entries:
 * @tree:
 * @key:
 * @error:
 *
 * Return value: the list of all entries in @key or NULL if none or error.
 **/
GList*
xprefs_tree_all_entries (XPrefsTree *tree,
                         const gchar *key,
                         GError **error)
{
  gchar **components;
  XPrefsDir *dir;

  /* treat special case */
  if (G_UNLIKELY (key[0] == '/' && key[1] == '\0'))
    return xprefs_dir_get_entries (tree->root);

  components = xprefs_split_path (key);
  if (G_LIKELY (components != NULL))
    {
      dir = xprefs_dir_lookup_subdir (tree->root, components, FALSE, error);
      g_strfreev (components);

      return xprefs_dir_get_entries (dir);
    }

  g_set_error (error, XPREFS_ERROR, XPREFS_ERROR_BAD_KEY,
               _("Invalid key '%s'"), key);

  return NULL;
}

/**
 * xprefs_tree_all_subdirs:
 * @tree:
 * @key:
 * @error:
 *
 * Return value: the list of all sub directories in @key or NULL if none
 *               or error.
 **/
GList*
xprefs_tree_all_subdirs (XPrefsTree *tree,
                         const gchar *key,
                         GError **error)
{
  gchar **components;
  XPrefsDir *dir;

  /* treat special case */
  if (G_UNLIKELY (key[0] == '/' && key[1] == '\0'))
    return xprefs_dir_get_subdirs (tree->root);

  components = xprefs_split_path (key);
  if (G_LIKELY (components != NULL))
    {
      dir = xprefs_dir_lookup_subdir (tree->root, components, FALSE, error);
      g_strfreev (components);

      return xprefs_dir_get_subdirs (dir);
    }

  g_set_error (error, XPREFS_ERROR, XPREFS_ERROR_BAD_KEY,
               _("Invalid key '%s'"), key);

  return NULL;
}


static void
xprefs_tree_foreach_internal (XPrefsDir *dir,
                              XPrefsTreeForeachFunc func,
                              gpointer user_data)
{
  GList *l;

  /* the recursion first... */
  for (l = xprefs_dir_get_subdirs (dir); l != NULL; l = l->next)
    xprefs_tree_foreach_internal (XPREFS_DIR (l->data), func, user_data);

  /* now, for each entry... */
  for (l = xprefs_dir_get_entries (dir); l != NULL; l = l->next)
    (*func) (XPREFS_ENTRY (l->data), user_data);
}

/**
 * xprefs_tree_foreach:
 * @tree:
 * @func:
 * @user_data:
 **/
void
xprefs_tree_foreach (XPrefsTree *tree,
                     XPrefsTreeForeachFunc func,
                     gpointer user_data)
{
  xprefs_tree_foreach_internal (tree->root, func, user_data);
}


