/* $Id$ */
/*-
 * Copyright (c) 2003, 2004 Benedikt Meurer <benny@xfce.org>
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <xprefs/xprefs-dbus.h>


/**
 * dbus_message_iter_get_xprefs_entry:
 * @iter: the iterator.
 *
 * Return value: new XPrefsEntry object or NULL on error.
 **/
XPrefsEntry*
dbus_message_iter_get_xprefs_entry (DBusMessageIter *iter)
{
  XPrefsEntry *entry = NULL;
  XPrefsValue *value;
  gchar *name;

  if (dbus_message_iter_get_arg_type (iter) != DBUS_TYPE_STRING)
    return NULL;

  name = dbus_message_iter_get_string (iter);
  dbus_message_iter_next (iter);

  if (dbus_message_iter_get_arg_type (iter) == DBUS_TYPE_NIL)
    value = NULL;
  else
    value = dbus_message_iter_get_xprefs_value (iter);

  entry = xprefs_entry_new_nocopy (g_strdup (name), value);

  dbus_free (name);

  return entry;
}

/**
 * dbus_message_iter_get_xprefs_value:
 * @iter: the iterator
 *
 * Return value: new XPrefsValue object or NULL on error.
 **/
XPrefsValue*
dbus_message_iter_get_xprefs_value (DBusMessageIter *iter)
{
  XPrefsValue *value;

  switch (dbus_message_iter_get_arg_type (iter))
    {
    case DBUS_TYPE_BOOLEAN:
      value = xprefs_value_new (XPREFS_VALUE_BOOL);
      xprefs_value_set_bool (value, dbus_message_iter_get_boolean (iter));
      break;

    case DBUS_TYPE_DOUBLE:
      value = xprefs_value_new (XPREFS_VALUE_DOUBLE);
      xprefs_value_set_double (value, dbus_message_iter_get_double (iter));
      break;

    case DBUS_TYPE_INT32:
      value = xprefs_value_new (XPREFS_VALUE_INT);
      xprefs_value_set_int (value, dbus_message_iter_get_int32 (iter));
      break;

    case DBUS_TYPE_STRING:
      {
        char *data;

        value = xprefs_value_new (XPREFS_VALUE_STRING);
        data = dbus_message_iter_get_string (iter);
        xprefs_value_set_string (value, data);
        dbus_free (data);
      }
      break;

    default:
      return NULL;
    }

  return value;
}

/**
 * dbus_message_iter_append_xprefs_entry:
 * @iter: the append iterator.
 * @entry:
 *
 * Return value: TRUE on success.
 */
dbus_bool_t
dbus_message_iter_append_xprefs_entry (DBusMessageIter *iter,
                                       const XPrefsEntry *entry)
{
  const XPrefsValue *value;
  const gchar *name;
  dbus_bool_t result;

  g_return_val_if_fail (entry != NULL, FALSE);

  name = xprefs_entry_get_name (entry);
  value = xprefs_entry_get_value (entry);

  result = dbus_message_iter_append_string (iter, name);

  if (result)
    {
      if (value != NULL)
        result = dbus_message_iter_append_xprefs_value (iter, value);
      else
        result = dbus_message_iter_append_nil (iter);
    }

  return result;
}

/**
 * dbus_message_iter_append_xprefs_value:
 * @iter:
 * @value:
 *
 * Return value: TRUE on success.
 **/
dbus_bool_t
dbus_message_iter_append_xprefs_value (DBusMessageIter *iter,
                                       const XPrefsValue *value)
{
  g_return_val_if_fail (value != NULL, FALSE);
  g_return_val_if_fail (XPREFS_VALUE_TYPE_VALID (value->type), FALSE);

  switch (value->type)
    {
    case XPREFS_VALUE_BOOL:
      {
        dbus_bool_t data;

        data = xprefs_value_get_bool (value);
        return dbus_message_iter_append_boolean (iter, data);
      }
      break;

    case XPREFS_VALUE_DOUBLE:
      {
        double data;

        data = xprefs_value_get_double (value);
        return dbus_message_iter_append_double (iter, data);
      }
      break;

    case XPREFS_VALUE_INT:
      {
        dbus_int32_t data;

        data = xprefs_value_get_int (value);
        return dbus_message_iter_append_int32 (iter, data);
      }
      break;

    case XPREFS_VALUE_STRING:
      {
        const char *data;

        data = xprefs_value_get_string (value);
        return dbus_message_iter_append_string (iter, data);
      }
      break;

    default:
      g_assert_not_reached ();
    }

  return FALSE;
}


/**
 * dbus_message_append_xprefs_value:
 * @message:
 * @value:
 *
 * Return value:
 */
dbus_bool_t
dbus_message_append_xprefs_value (DBusMessage *message,
                                  const XPrefsValue *value)
{
  DBusMessageIter iter;
  dbus_message_append_iter_init (message, &iter);
  return dbus_message_iter_append_xprefs_value (&iter, value);
}


/**
 * dbus_connection_send_and_process_reply:
 * @connection:
 * @message:
 * @error:
 *
 * Automatically processes the returned reply, that is, if the reply is an
 * error message, this error is translated into an error object and stored
 * in  @error.
 *
 * Return value: reply or NULL on error.
 **/
DBusMessage*
dbus_connection_send_and_process_reply (DBusConnection *connection,
                                        DBusMessage *message,
                                        DBusError *error)
{
  DBusMessage *reply;

  reply = dbus_connection_send_with_reply_and_block (connection,
                                                     message,
                                                     -1,
                                                     error);

  /* check if the reply is an error message */
  if (reply != NULL && dbus_set_error_from_message (error, reply))
    {
      dbus_message_unref (reply);

      return NULL;
    }

  return reply;
}

