/* $Id$ */
/*-
 * Copyright (c) 2003, 2004 Benedikt Meurer <benny@xfce.org>
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <xprefs/xprefs-error.h>
#include <xprefs/xprefs-private.h>
#include <xprefs/xprefs-value.h>


/**
 * xprefs_value_type_to_string:
 * @type:
 **/
const gchar*
xprefs_value_type_to_string (XPrefsValueType type)
{
  switch (type)
    {
    case XPREFS_VALUE_BOOL:   return "bool";
    case XPREFS_VALUE_DOUBLE: return "double";
    case XPREFS_VALUE_INT:    return "int";
    case XPREFS_VALUE_STRING: return "string";
    case XPREFS_VALUE_INVALID:return "*invalid*";
    }

  g_assert_not_reached ();
  return NULL;
}

/**
 * xprefs_value_type_from_string:
 * @type_str:
 **/
XPrefsValueType
xprefs_value_type_from_string (const gchar *type_str)
{
  if (strcmp (type_str, "bool") == 0)         return XPREFS_VALUE_BOOL;
  else if (strcmp (type_str, "double") == 0)  return XPREFS_VALUE_DOUBLE;
  else if (strcmp (type_str, "int") == 0)     return XPREFS_VALUE_INT;
  else if (strcmp (type_str, "string") == 0)  return XPREFS_VALUE_STRING;
  else                                        return XPREFS_VALUE_INVALID;
}


typedef struct _XPrefsRealValue XPrefsRealValue;

struct _XPrefsRealValue
{
  XPrefsValueType type;
  union {
    gboolean      bool_data;
    gdouble       double_data;
    gint32        int_data;
    gchar        *string_data;
  } d;
};

#define REAL_VALUE(value) ((XPrefsRealValue *)(void *)(value))


XPrefsValue*
xprefs_value_new (XPrefsValueType type)
{
  XPrefsRealValue *real;

  real        = g_new0 (XPrefsRealValue, 1);
  real->type  = type;

  return (XPrefsValue *)real;
}

XPrefsValue*
xprefs_value_new_from_string (XPrefsValueType type,
                              const gchar *value_str,
                              GError **error)
{
  XPrefsValue *value;

  g_return_val_if_fail (XPREFS_VALUE_TYPE_VALID (type), NULL);
  g_return_val_if_fail (value_str != NULL, NULL);

  value = xprefs_value_new (type);

  switch (type)
    {
    case XPREFS_VALUE_BOOL:
      switch (g_ascii_tolower (*value_str))
        {
        case 't':
        case '1':
        case 'y':
          xprefs_value_set_bool (value, TRUE);
          break;

        case 'f':
        case '0':
        case 'n':
          xprefs_value_set_bool (value, FALSE);
          break;

        default:
          g_set_error (error, XPREFS_ERROR, XPREFS_ERROR_PARSE_ERROR,
                       _("Didn't understand '%s' (expected bool)"),
                       value_str);
          xprefs_value_free (value);
          value = NULL;
          break;
        }
      break;

    case XPREFS_VALUE_DOUBLE:
      {
        gchar *endptr = NULL;
        gdouble data;

        data = g_ascii_strtod (value_str, &endptr);
        if (endptr == value_str)
          {
            g_set_error (error, XPREFS_ERROR, XPREFS_ERROR_PARSE_ERROR,
                         _("Didn't understand '%s' (expected double)"),
                         value_str);
            xprefs_value_free (value);
            value = NULL;
          }
        else if (errno == ERANGE)
          {
            g_set_error (error, XPREFS_ERROR, XPREFS_ERROR_PARSE_ERROR,
                         _("Double '%s' is too large or too small"),
                         value_str);
            xprefs_value_free (value);
            value = NULL;
          }
        else
          xprefs_value_set_double (value, data);
      }
      break;

    case XPREFS_VALUE_INT:
      {
        gchar *endptr = NULL;
        long data;

        data = strtol (value_str, &endptr, 10);
        if (endptr == value_str)
          {
            g_set_error (error, XPREFS_ERROR, XPREFS_ERROR_PARSE_ERROR,
                         _("Didn't understand '%s' (expected integer)"),
                         value_str);
            xprefs_value_free (value);
            value = NULL;
          }
        else if (errno == ERANGE || data > G_MAXINT32 || data < G_MININT32)
          {
            g_set_error (error, XPREFS_ERROR, XPREFS_ERROR_PARSE_ERROR,
                         _("Integer '%s' is too large or too small"),
                         value_str);
            xprefs_value_free (value);
            value = NULL;
          }
        else
          xprefs_value_set_int (value, (gint32)data);
      }
      break;

    case XPREFS_VALUE_STRING:
      if (!g_utf8_validate (value_str, -1, NULL))
        {
          g_set_error (error, XPREFS_ERROR, XPREFS_ERROR_PARSE_ERROR,
                       _("Text contains invalid UTF-8"));
          xprefs_value_free (value);
          value = NULL;
        }
      else
        {
          xprefs_value_set_string (value, value_str);
        }
      break;

    default:
      g_assert_not_reached ();
    }

  return value;
}

XPrefsValue*
xprefs_value_copy (const XPrefsValue *value)
{
  XPrefsRealValue *real;

  real = g_memdup (value, sizeof (*real));

  /* strings need to be duplicated */
  if (real->type == XPREFS_VALUE_STRING)
    real->d.string_data = g_strdup (real->d.string_data);

  return (XPrefsValue *)real;
}

void
xprefs_value_free (XPrefsValue *value)
{
  XPrefsRealValue *real;

  g_return_if_fail (value != NULL);

  real = REAL_VALUE (value);
  if (value->type == XPREFS_VALUE_STRING && real->d.string_data != NULL)
    g_free (real->d.string_data);
  g_free (real);
}

gchar*
xprefs_value_to_string (const XPrefsValue *value)
{
  gchar *retval = NULL;

  g_return_val_if_fail (value != NULL, NULL);
  g_return_val_if_fail (XPREFS_VALUE_TYPE_VALID (value->type), NULL);

  switch (value->type)
    {
    case XPREFS_VALUE_BOOL:
      retval = g_strdup (xprefs_value_get_bool (value) ? "true" : "false");
      break;

    case XPREFS_VALUE_DOUBLE:
      retval = g_new0 (gchar, G_ASCII_DTOSTR_BUF_SIZE + 1);
      retval = g_ascii_dtostr (retval, G_ASCII_DTOSTR_BUF_SIZE + 1,
                               xprefs_value_get_double (value));
      break;

    case XPREFS_VALUE_INT:
      retval = g_strdup_printf ("%ld", (long)xprefs_value_get_int (value));
      break;

    case XPREFS_VALUE_STRING:
      retval = g_strdup (xprefs_value_get_string (value));
      break;

    default:
      g_assert_not_reached ();
    }

  return retval;
}

gboolean
xprefs_value_get_bool (const XPrefsValue *value)
{
  g_return_val_if_fail (value != NULL, FALSE);
  g_return_val_if_fail (value->type == XPREFS_VALUE_BOOL, FALSE);

  return REAL_VALUE (value)->d.bool_data;
}

gdouble
xprefs_value_get_double (const XPrefsValue *value)
{
  g_return_val_if_fail (value != NULL, 0.0);
  g_return_val_if_fail (value->type == XPREFS_VALUE_DOUBLE, 0.0);

  return REAL_VALUE (value)->d.double_data;
}

gint32
xprefs_value_get_int (const XPrefsValue *value)
{
  g_return_val_if_fail (value != NULL, 0);
  g_return_val_if_fail (value->type == XPREFS_VALUE_INT, 0);

  return REAL_VALUE (value)->d.int_data;
}

const gchar*
xprefs_value_get_string (const XPrefsValue *value)
{
  g_return_val_if_fail (value != NULL, NULL);
  g_return_val_if_fail (value->type == XPREFS_VALUE_STRING, NULL);

  return REAL_VALUE (value)->d.string_data;
}


/**
 * xprefs_value_steal_string:
 * @value: an XPrefsValue object.
 *
 * "Unlinks" the string from @value and returns it.
 *
 * Return value: the string.
 **/
gchar*
xprefs_value_steal_string (XPrefsValue *value)
{
  gchar *string;

  g_return_val_if_fail (value != NULL, NULL);
  g_return_val_if_fail (value->type == XPREFS_VALUE_STRING, NULL);

  string = REAL_VALUE (value)->d.string_data;
  REAL_VALUE (value)->d.string_data = NULL;

  return string;
}


void
xprefs_value_set_bool (XPrefsValue *value, gboolean bool_data)
{
  g_return_if_fail (value != NULL);
  g_return_if_fail (value->type == XPREFS_VALUE_BOOL);

  REAL_VALUE (value)->d.bool_data = bool_data;
}

void
xprefs_value_set_double (XPrefsValue *value, gdouble double_data)
{
  g_return_if_fail (value != NULL);
  g_return_if_fail (value->type == XPREFS_VALUE_DOUBLE);

  REAL_VALUE (value)->d.double_data = double_data;
}

void
xprefs_value_set_int (XPrefsValue *value, gint32 int_data)
{
  g_return_if_fail (value != NULL);
  g_return_if_fail (value->type == XPREFS_VALUE_INT);

  REAL_VALUE (value)->d.int_data = int_data;
}

void
xprefs_value_set_string (XPrefsValue *value, const gchar *string_data)
{
  xprefs_value_set_string_nocopy (value, g_strdup (string_data));
}

void
xprefs_value_set_string_nocopy (XPrefsValue *value, gchar *string_data)
{
  g_return_if_fail (value != NULL);
  g_return_if_fail (value->type == XPREFS_VALUE_STRING);

  /* free old value first */
  if (G_LIKELY (REAL_VALUE (value)->d.string_data != NULL))
    g_free (REAL_VALUE (value)->d.string_data);

  REAL_VALUE (value)->d.string_data = string_data;
}
